/* queryForm.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "queryForm.h"
#include "database.h"
#include "playlistDialog.h"
#include "playlist.h"
#include "tableFieldKey.h"
#include "tableField.h"
#include "formDef.h"
#include "srandObject.h"
#include "string_utils.h"
#include "busyMessage.h"
#include "infoMessage.h"

#include <map>
#include <set>
#include <algorithm>

QueryForm::
QueryForm(const Dataset &data, int form_id) :
  PalomaDialog(data),
  _form_id(form_id),
  _ok_button("OK"),
  _build_button("Build"),
  _cancel_button("Cancel")
{
  if (_form_id >= 0) {
    _form.layout_form(_form_id, _data);
    _form.show();
  }
  _playlist_form_id = get_related_form_id(_form_id, TI_songs);

  layout_window();
  setup();
}

QueryForm::
~QueryForm() {
}

void QueryForm::
ok_button_clicked() {
  hide();
  build_button_clicked();
  delete_self();
}

void QueryForm::
build_button_clicked() {
  _form.update();
  build_playlist();
}

void QueryForm::
cancel_button_clicked() {
  delete_self();
}

void QueryForm::
layout_window() {
  set_title(_form.get_def()._name);

  vbox_.pack_start(_form, true, true, 0);

  // The OK button.
  _ok_button.set_usize(80, 36);
  action_area_.pack_start(_ok_button, true, false, 0);
  _ok_button.set_flags(GTK_CAN_DEFAULT);
  _ok_button.grab_default();
  _ok_button.show();
  connect_to_method(_ok_button.clicked, this, &ok_button_clicked);
  
  // The Build button.
  _build_button.set_usize(80, 36);
  action_area_.pack_start(_build_button, true, false, 0);
  _build_button.set_flags(GTK_CAN_DEFAULT);
  _build_button.show();
  connect_to_method(_build_button.clicked, this, &build_button_clicked);
  
  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}

void QueryForm::
build_playlist() {
  BusyMessage bm("Building playlist");

  VariableDef vars;
  _form.get_variables(vars);

  string sql_command = 
    trim_blanks(expand_variables(_form.get_def()._general_sql_command,
				 vars));

  if (sql_command.empty()) {
    sql_command = "SELECT song_id, performer, name FROM songs";
  }

  if (!_form._ratio_slider_values.empty()) {
    // If we have some ratio sliders on the form, apply them now.
    compute_ratio_effect(sql_command);

  } else {
    // Otherwise, just issue the select and build up the playlist.
    QueryResult qr = music_db->sql_command(sql_command, true);
    int num_rows = qr.get_num_rows();
    if (num_rows == 0) {
      if (qr.success()) {
	new InfoMessage("No items were returned.");
      } else {
	new InfoMessage(qr.get_error_message());
      }

    } else {
      Playlist *playlist = new Playlist;
      playlist->append_data(qr);
      new PlaylistDialog(_data, _playlist_form_id, playlist);
    }   
  }
}

void QueryForm::
compute_ratio_effect(const string &sql_command) {
  // Identify all the sliders that have a nonzero value.  These will
  // determine the likelihood of choosing a particular on/off
  // attribute.

  // We'll set up all the nonzero sliders in a row.  Each slider
  // indicates the proportional chance to select a song with that
  // attribute.  We position all the sliders on a number line, and
  // then throw darts at that number line--when we pick a number
  // between slider position [n-1] and slider position [n], that will
  // have selected attribute [n].

  typedef ::map<double, TableFieldKey> Sliders;
  Sliders sliders;
  double net_value = 0.0;

  // First, identify all the sliders on the form, and line them up on
  // the number line.
  Form::RatioSliderValues::const_iterator rsvi;
  for (rsvi = _form._ratio_slider_values.begin();
       rsvi != _form._ratio_slider_values.end();
       ++rsvi) {
    double value = (*rsvi).second;
    if (value != 0.0) {
      net_value += value;
      sliders.insert(Sliders::value_type(net_value, (*rsvi).first));
    }
  }

  int num_sliders = sliders.size();
  assert(num_sliders != 0);

  // Now to choose songs, we'll be throwing darts at the number line
  // to pick sliders according to their relative chance.  For each
  // slider that comes up, we'll want to randomly choose a song that
  // has the corresponding attribute.  Thus, we need to prepare a
  // bucket for each slider, filled with all the songs (or a random
  // subsample) that match the associated attribute.

  typedef vector<int> SongIDs;
  typedef ::map<TableFieldKey, SongIDs> Buckets;
  Buckets buckets;

  Sliders::const_iterator sli;
  for (sli = sliders.begin(); sli != sliders.end(); ++sli) {
    TableFieldKey key = (*sli).second;
    const TableField &field = key.get_field();

    QueryResult qr;
    SongIDs &ids = buckets[key];

    if (key._table_id == TI_songs) {
      qr = music_db->sql_command
	("SELECT song_id FROM songs "
	 "WHERE " + field._field_name + "=" + DataCellBool(true));
    } else if (key._table_id == TI_songs_extension) {
      qr = music_db->sql_command
	("SELECT song_id FROM songs_extension "
	 "WHERE " + field._field_name + "=" + DataCellBool(true));
    } else {
      new InfoMessage("Currently unable to build playlist from table " +
		      get_table_name(key._table_id));
      return;
    }
    int num_rows = qr.get_num_rows();
    for (int i = 0; i < num_rows; i++) {
      ids.push_back(qr.get_cell(i, 0).get_int());
    }

    // Now shuffle the list of id's into random order.
    SRandObject srand;
    random_shuffle(ids.begin(), ids.end(), srand);
  }

  // Ok, now we have a bunch of buckets of integer song ID's, and a
  // bunch of sliders on a number line.  Start throwing darts.

  set<int> used_ids;
  Playlist *playlist = new Playlist(TI_songs);

  playlist->_columns.push_back(Playlist::Column("performer", DataCell::DC_int));
  playlist->_columns.push_back(Playlist::Column("name", DataCell::DC_string));
  playlist->_columns.push_back(Playlist::Column("Length", DataCell::DC_elapsed_time));
  playlist->_columns.push_back(Playlist::Column("Type", DataCell::DC_string));


  bool okflag = true;
  while (okflag) {
    double dart_point = 
      net_value * (double)rand() / ((double)RAND_MAX + 1.0);
    
    sli = sliders.lower_bound(dart_point);
    // Since dart_point is always less than net_value, there will always
    // be an entry in the map greater than dart_point.  So this should
    // never return end().
    assert(sli != sliders.end());

    TableFieldKey key = (*sli).second;
    const TableField &field = key.get_field();
    SongIDs &ids = buckets[key];

    int song_id;
    do {
      if (ids.empty()) {
	// We just tried to request a song from a bucket that had no more
	// songs to give.  We're done.
	okflag = false;
      } else {
      // Now pull the next element out of the indicated bucket.
	song_id = ids.back();
	ids.pop_back();
      }

      // Keep trying to pull out an ID until we've got one we haven't
      // seen before.  (We have to do this because a given song might
      // appear in several buckets.)

    } while (okflag && used_ids.find(song_id) != used_ids.end());
    
    if (okflag) {
      Playlist::Row row(song_id);

      TableDataProxy song(TI_songs, song_id);
      row._column_values.push_back(song[SFI_performer]);
      row._column_values.push_back(song[SFI_name]);
      row._column_values.push_back(song[SFI_length]);
      row._column_values.push_back(DataCell(field._label));

      playlist->_rows.push_back(row);
      used_ids.insert(song_id);
    }
  }

  new PlaylistDialog(_data, _playlist_form_id, playlist);
}
